DISCLOSURE_SYSTEM_NT16 = 'NT16'
DISCLOSURE_SYSTEM_NT17 = 'NT17'
DISCLOSURE_SYSTEM_NT18 = 'NT18'
DISCLOSURE_SYSTEM_NT19 = 'NT19'
DISCLOSURE_SYSTEM_NT20 = 'NT20'
DISCLOSURE_SYSTEM_NL_INLINE_2024 = 'NL-INLINE-2024'
DISCLOSURE_SYSTEM_NL_INLINE_2024_GAAP_OTHER = 'NL-INLINE-2024-GAAP-OTHER'
DISCLOSURE_SYSTEM_NL_INLINE_2025 = 'NL-INLINE-2025'
DISCLOSURE_SYSTEM_NL_INLINE_2025_GAAP_OTHER = 'NL-INLINE-2025-GAAP-OTHER-PREVIEW'
DISCLOSURE_SYSTEM_NL_INLINE_MULTI_TARGET = 'NL-INLINE-MULTI-TARGET'

DISCLOSURE_SYSTEM_YEARS = {
    DISCLOSURE_SYSTEM_NL_INLINE_2024: 2024,
    DISCLOSURE_SYSTEM_NL_INLINE_2024_GAAP_OTHER: 2024,
    DISCLOSURE_SYSTEM_NL_INLINE_2025: 2025,
    DISCLOSURE_SYSTEM_NL_INLINE_2025_GAAP_OTHER: 2025,
}

ALL_NL_INLINE_DISCLOSURE_SYSTEMS = [
    DISCLOSURE_SYSTEM_NL_INLINE_2024,
    DISCLOSURE_SYSTEM_NL_INLINE_2024_GAAP_OTHER,
    DISCLOSURE_SYSTEM_NL_INLINE_2025,
    DISCLOSURE_SYSTEM_NL_INLINE_2025_GAAP_OTHER,
]

NL_INLINE_DISCLOSURE_SYSTEMS_2025_AND_NEWER = [
    DISCLOSURE_SYSTEM_NL_INLINE_2025,
    DISCLOSURE_SYSTEM_NL_INLINE_2025_GAAP_OTHER,
]

NL_INLINE_GAAP_IFRS_DISCLOSURE_SYSTEMS = [
    DISCLOSURE_SYSTEM_NL_INLINE_2024,
    DISCLOSURE_SYSTEM_NL_INLINE_2025
]

NL_INLINE_GAAP_OTHER_DISCLOSURE_SYSTEMS = [
    DISCLOSURE_SYSTEM_NL_INLINE_2024_GAAP_OTHER,
    DISCLOSURE_SYSTEM_NL_INLINE_2025_GAAP_OTHER,
]

NL_INLINE_GAAP_OTHER_DISCLOSURE_SYSTEMS_2025_AND_NEWER = [
    DISCLOSURE_SYSTEM_NL_INLINE_2025_GAAP_OTHER,
]

NL_INLINE_MULTI_TARGET_DISCLOSURE_SYSTEMS = [
    DISCLOSURE_SYSTEM_NL_INLINE_MULTI_TARGET,
]

NL_INLINE_OTHER_DISCLOSURE_SYSTEMS = [
    DISCLOSURE_SYSTEM_NL_INLINE_2024_GAAP_OTHER,
    DISCLOSURE_SYSTEM_NL_INLINE_2025_GAAP_OTHER,
    DISCLOSURE_SYSTEM_NL_INLINE_MULTI_TARGET,
]

NT_DISCLOSURE_SYSTEMS = [
    DISCLOSURE_SYSTEM_NT16,
    DISCLOSURE_SYSTEM_NT17,
    DISCLOSURE_SYSTEM_NT18,
    DISCLOSURE_SYSTEM_NT19,
    DISCLOSURE_SYSTEM_NT20,
]
