ARL_2022_PREVIEW = "arl-2022-preview"
ARL_2024_PREVIEW = "arl-2024-preview"
ARL_2025_PREVIEW = "arl-2025-preview"
ARL_2024_MULTI_TARGET_PREVIEW = "arl-2024-multi-target-preview"
ARL_2025_MULTI_TARGET_PREVIEW = "arl-2025-multi-target-preview"
DKFIN_2024_MULTI_TARGET_PREVIEW = "dkfin-2024-multi-target-preview"

ALL_DISCLOSURE_SYSTEMS = [
    ARL_2022_PREVIEW,
    ARL_2024_PREVIEW,
    ARL_2025_PREVIEW,
    ARL_2024_MULTI_TARGET_PREVIEW,
    ARL_2025_MULTI_TARGET_PREVIEW,
    DKFIN_2024_MULTI_TARGET_PREVIEW
]

ARL_DISCLOSURE_SYSTEMS = [
    ARL_2022_PREVIEW,
    ARL_2024_PREVIEW,
    ARL_2025_PREVIEW,
    ARL_2024_MULTI_TARGET_PREVIEW,
    ARL_2025_MULTI_TARGET_PREVIEW
]

STAND_ALONE_DISCLOSURE_SYSTEMS = [
    ARL_2022_PREVIEW,
    ARL_2024_PREVIEW,
    ARL_2025_PREVIEW
]

ARL_MULTI_TARGET_DISCLOSURE_SYSTEMS = [
    ARL_2024_MULTI_TARGET_PREVIEW,
    ARL_2025_MULTI_TARGET_PREVIEW
]

DKFIN_MULTI_TARGET_DISCLOSURE_SYSTEMS = [
    DKFIN_2024_MULTI_TARGET_PREVIEW
]
